<copyright> MU unit.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Unit.t,v 1.21 1999/03/09 15:18:26 tiggr Exp $</id>
    </copyright>

implementation class
Unit: Common
{
  <doc> All units ever read.  </doc>
  static MutableDictionary units;

  <doc> The {_builtin} unit, i.e. the top unit.  </doc>
  public static Unit builtin_unit;

  <doc> The search mark used to mark units already visited during
      traversal.  </doc>
  static int search_mark;
}

redefine void
  load Array arguments
{
  units = [MutableDictionary new];
  builtin_unit = [[self alloc] init "_builtin_" directory "_builtin_/"];
}

<doc> The designated way of getting a unit.  If it has not yet been
    loaded, and the optional {load_p} is {TRUE}, it will be loaded, plus
    anything upon which it depends.  Adjusts {all_user_units} and
    {all_units}.  </doc>
instance (id)
  named String name
  load: boolean load_p = NO
{
  Unit u = units[name];

  if (!u && load_p)
    {
      String leaf_name = [[MutableByteString new] print (name, ".u")];
      File f;

      catch (self)
	bind ((file-error,
	       {
		 [[[Parser error] print ([[condition object] name],
					 ": ", [condition message])] nl];
		 [self throw nil];
		 nil;
	       }))
	  ({
	    f = [File open leaf_name alongPath [Parser path] subdir: name];
	    if (!f)
	      [[[Parser error] print (leaf_name, ": unable to locate file")]
		nl];
	    else
	      {
		u = [[self alloc] init name directory [f directoryName]];

		LexUnit l = [[LexUnit alloc] initWithStream f named [f name]];
		ParseUnit p = [[ParseUnit alloc] init l unit u];
      
		[p file];
	      }
	  });

       [all_user_units add u];
       [all_units add u];
    }

  = u;
}

<doc> Return the {Mapped} {Collection} of units.  </doc>
Mapped
  units
{
  = units;
}

end;

implementation instance
Unit
{
  <doc> The name of this unit, sans extension.  </doc>
  public String name;

  <doc> The directory containing this unit.  </doc>
  public String directory;

  <doc> The other units this unit depends on.  </doc>
  MutableArray used;

  <doc> The files in this unit.  </doc>
  MutableDictionary files;

  <doc> The metas defined in this unit, such that the instance {U.M} is in
      the {metas} of the unit {U}.  </doc>
  MutableDictionary metas;

  <doc> The types when a file of this unit is read.  </doc>
  MutableSet unit_types;

  <doc> The mark used when searching in units.  </doc>
  int mark;
}

<doc> Designated initializer.  </doc>
id
       init String n
  directory String d
{
  (name, directory) = (n, d);
  units[name] = self;
  metas = [MutableDictionary new];
  files = [MutableDictionary new];

  = self;
}

void
  addMeta Meta m
{
  metas[[m name]] = m;
}

<doc> Return the file named {name}.  </doc>
TomFile
  fileNamed String name
{
  = files[name];
}

<doc> Designated way to obtain a meta.  Iff {!rec_p}, only the receiving
    unit will be searched.  Otherwise, all units transitively being
    depended upon are also searched.  Upon multiple meta's with the same
    name being found, it is not an error if the receiving unit defines
    one, since that one will be returned.  </doc>
Meta
  metaNamed String meta_name
   recurse: boolean rec_p = NO
{
  if (!rec_p)
    return metas[meta_name];

  boolean multiple;
  Meta meta;

  (multiple, meta) = [self metaNamed meta_name found nil mark ++search_mark];

  = multiple ? metas[meta_name] : meta;
}

<doc> Search for the meta named {meta_name}, using the mark {k} to prevent
    multiple visits.  {found} is the meta already found.  </doc>
private (boolean, Meta)
  metaNamed String meta_name
      found Meta found
       mark int k
{
  if (mark == k)
    return (NO, nil);
  mark = k;

  Meta this_found, m = metas[meta_name];

  if (m != nil)
    return !found ? (NO, m) : (YES, nil);

  if (used != nil)
    {
      int i, n = [used length];

      for (i = 0; i < n; i++)
	{
	  boolean multiple;

	  (multiple, m) = [used[i] metaNamed meta_name found found mark mark];
	  if (multiple)
	    return (YES, nil);
	  if (m != nil)
	    {
	      if (found != nil || this_found != nil)
		return (YES, nil);
	      this_found = m;
	    }
	}
    }

  = (NO, this_found);
}

<doc> Return the {Mapped} collection of {files}.  </doc>
Mapped
  files
{
  = files;
}

<doc> Return the {Mapped} collection of {metas}.  </doc>
Mapped
  metas
{
  = metas;
}

<doc> Return the {unit_types}, creating it if necessary.  </doc>
MutableSet
  types
{
  if (!unit_types)
    {
      unit_types = [MutableSet new];
      [BasicType addTypes unit_types];
      [self types unit_types mark ++search_mark];
    }
  = unit_types;
}

<doc> Add the names of the metas from the receiving unit, and any unit
    that it uses, to the {set}.  </doc>
void
  types MutableSet set
   mark int k
{
  if (mark == k)
    return;
  mark = k;

  Enumerator e = [metas keyEnumerator];
  String s;
  while ({boolean b; (b, s) = [e next]; b;})
    [set add s];

  if (used != nil)
    {
      int i, n = [used length];

      for (i = 0; i < n; i++)
	[used[i] types set mark k];
    }
}

<doc> Return the used units.  This is {nil} if this unit uses no other
    units.  </doc>
Indexed
  used
{
  = used;
}

<doc> Emit the {name}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" name=", name)];
}

/********** interface to LexUnit **********/

<doc> Create and return a new file contained in this unit.  </doc>
TomFile
  fileWithName String name
{
  TomFile f = [TomFile named name inUnit self];

  files[[f name]] = f;
  = f;
}

<doc> Called by the lexer to inform us of another unit we depend on.  </doc>
void
  uses String name
{
  id u = [isa named name load: YES];

  if (u != nil)
    {
      if (!used)
	used = [MutableObjectArray new];
      [used add u];
    }
}

end;
